import re
from datetime import timedelta
from pathlib import Path

from pymediainfo import MediaInfo

from auto_qpf.enums import ChapterType


class ChapterGenerator:
    __slots__ = ()

    def generate_ogm_chapters(
        self,
        media_info_obj: MediaInfo,
        output_path: Path,
        chapter_chunks: float = 5.0,
        extract_tagged: bool = True,
        extract_named: bool = True,
        extract_numbered: bool = True,
        write_to_file: bool = True,
    ):
        """
        Detect rather or not input file has numbered, named, tagged chapters or no chapters.

        If needed this will generate OGM/txt based chapters in the format:
        Chapter 01
        Chapter 02
        etc...

        Args:
            media_info_obj (MediaInfo): MediaInfo.parse() object.
            output_path (Path): Output path for the *.txt chapter file.
            chapter_chunks (float): Chunks to create chapters automatically in percentage.
            Defaults to 5.0.
            If a movie is an hour and a half long this would give you chapters about every 11 minutes.
            extract_tagged (bool, optional): If tagged chapters are accepted and the input has them extract those.
            Defaults to True.
            extract_named (bool, optional): If named chapters are accepted and the input has them extract those.
            Defaults to True.
            extract_numbered (bool, optional): If numbered chapters are accepted (in the correct format) and the input has them extract those.
            Defaults to True.
            write_to_file (bool, optional): If True, writes to file and returns Path. If False, returns chapter content as string.
            Defaults to True.
        """

        get_chapters = self._get_media_info_obj_chapters(media_info_obj)

        # if chapters was detected determine the type and if we want to keep them
        if get_chapters:
            chapter_type = self._determine_chapter_type(get_chapters)
            if chapter_type[0] == ChapterType.NAMED:
                if extract_named:
                    return self._extract_chapters(
                        chapter_type[0], chapter_type[1], output_path, write_to_file
                    )
                elif not extract_named:
                    generate_chapters = self._generate_chapters(
                        media_info_obj, chapter_chunks
                    )
                    return self._write_new_numbered_chapters(
                        generate_chapters, output_path, write_to_file
                    )

            if chapter_type[0] == ChapterType.NUMBERED:
                if extract_numbered:
                    try:
                        return self._extract_chapters(
                            chapter_type[0], chapter_type[1], output_path, write_to_file
                        )
                    except IndexError:
                        generate_chapters = self._generate_chapters(
                            media_info_obj, chapter_chunks
                        )
                        return self._write_new_numbered_chapters(
                            generate_chapters, output_path, write_to_file
                        )
                elif not extract_numbered:
                    generate_chapters = self._generate_chapters(
                        media_info_obj, chapter_chunks
                    )
                    return self._write_new_numbered_chapters(
                        generate_chapters, output_path, write_to_file
                    )

            if chapter_type[0] == ChapterType.TAGGED:
                if extract_tagged:
                    return self._extract_chapters(
                        chapter_type[0], chapter_type[1], output_path, write_to_file
                    )
                elif not extract_tagged:
                    generate_chapters = self._generate_chapters(
                        media_info_obj, chapter_chunks
                    )
                    return self._write_new_numbered_chapters(
                        generate_chapters, output_path, write_to_file
                    )

        # if no chapters was detected
        elif not get_chapters:
            generate_chapters = self._generate_chapters(media_info_obj, chapter_chunks)
            return self._write_new_numbered_chapters(
                generate_chapters, output_path, write_to_file
            )

    def _determine_chapter_type(self, media_info_menu: dict):
        menu = self._get_menu_info_only(media_info_menu)
        menu_str = " ".join(menu[2])
        menu_str_reversed = " ".join(reversed(menu[2]))

        if re.search(r"\d+:\d+:\d+\.\d+", menu_str):
            return ChapterType.TAGGED, menu[0]

        else:
            try:
                # check for numbered chapters
                start_match = re.search(r"chapter\s*(\d+)", menu_str, re.IGNORECASE)
                end_match = re.search(
                    r"chapter\s*(\d+)",
                    menu_str_reversed,
                    re.IGNORECASE,
                )

                if start_match and end_match:
                    chapters_start_numbered = start_match.group(1)
                    chapters_end_numbered = end_match.group(1)

                    return (
                        ChapterType.NUMBERED,
                        menu[0],
                        chapters_start_numbered,
                        chapters_end_numbered,
                    )
                else:
                    # if regex didn't match, treat as named chapters
                    return ChapterType.NAMED, menu[0]

            # if chapters are not numbered assume Named (since we check for tagged chapters already)
            except AttributeError:
                return ChapterType.NAMED, menu[0]

    def _generate_chapters(self, media_info: MediaInfo, chapter_chunks: float):
        duration_ms = int(media_info.general_tracks[0].duration) / 1000

        # Calculate the duration for each chapter
        chunk_percentage = chapter_chunks / 100
        chapter_interval = duration_ms * chunk_percentage

        chapter_dict = {1: "00:00:00.000"}

        # Calculate the number of chapters based on the chunk percentage
        num_chapters = int(1 / chunk_percentage)

        for i in range(2, num_chapters + 1):
            chapter_seconds = i * chapter_interval
            formatted_time = self._convert_to_time_format(chapter_seconds)
            # Round the seconds part to 3 decimal places
            formatted_time = ":".join(
                part.zfill(2) for part in formatted_time.split(":")
            )
            chapter_dict[i] = formatted_time

        return chapter_dict

    @staticmethod
    def _get_media_info_obj_chapters(media_info_obj: MediaInfo):
        menu_stream_count = media_info_obj.general_tracks[0].count_of_menu_streams
        if menu_stream_count and int(menu_stream_count) > 0:
            return media_info_obj.menu_tracks[0].to_data()
        return None

    @staticmethod
    def _get_menu_info_only(media_info_menu: dict):
        chapter_dict = {}
        chapters_only = list(media_info_menu.keys()).index("chapters_pos_end") + 1
        chapter_keys = list(media_info_menu.keys())[chapters_only:]
        chapter_values = list(media_info_menu.values())[chapters_only:]

        for val, key in zip(chapter_keys, chapter_values, strict=True):
            chapter_dict.update({val: key})

        return chapter_dict, chapter_keys, chapter_values

    @staticmethod
    def _write_new_numbered_chapters(
        chapter_dict: dict, output_path: Path, write_to_file: bool = True
    ):
        output_path = Path(output_path)

        lines = []
        for num, tag in enumerate(chapter_dict.keys(), start=1):
            num_str = str(num).zfill(2)
            tag_time = chapter_dict[tag]
            lines.append(
                f"CHAPTER{num_str}={tag_time}\nCHAPTER{num_str}NAME=Chapter {num_str}\n"
            )

        content = "".join(lines)

        if write_to_file:
            with open(output_path, "wt+", encoding="utf-8") as chapt_out:
                chapt_out.write(content)
            if output_path.is_file():
                return output_path
        else:
            return content

    @staticmethod
    def _extract_chapters(
        chapter_type: ChapterType,
        chapter_dict: dict,
        output_path: Path,
        write_to_file: bool = True,
    ):
        output_path = Path(output_path)

        # if we have numbered chapters we're going to check that they are in the correct
        # format of Chapters 01, Chapters 02, etc...
        if chapter_type == ChapterType.NUMBERED:
            first_chapter_num = list(chapter_dict.values())[0].split(" ")[1].strip()
            if first_chapter_num != "01":
                new_chapter_dict = {}
                for num, chap in enumerate(chapter_dict, start=1):
                    new_chapter_dict.update({chap: f"Chapter {str(num).zfill(2)}"})
                chapter_dict = new_chapter_dict

        # extract chapters
        lines = []
        for num, tag in enumerate(chapter_dict, start=1):
            num_str = str(num).zfill(2)
            l_tag = tag.replace("_", ":")[:-3]
            r_tag = tag[-3:]
            new_tag = f"{l_tag}.{r_tag}"
            language_pattern = re.compile(
                r"^[ :;?!_-]*(([a-zA-Z]{2,3}:){1,2})*[ :;?!_-]*"
            )
            value = re.sub(language_pattern, "", str(chapter_dict[tag])).strip()
            lines.append(f"CHAPTER{num_str}={new_tag}\nCHAPTER{num_str}NAME={value}\n")

        content = "".join(lines)

        if write_to_file:
            with open(output_path, "wt+", encoding="utf-8") as chapt_out:
                chapt_out.write(content)
            if output_path.is_file():
                return output_path
        else:
            return content

    @staticmethod
    def _convert_to_time_format(seconds):
        td = timedelta(seconds=seconds)
        # Format the timedelta with three decimal places for seconds
        formatted_time = str(td).rsplit(".", 1)[0] + f".{td.microseconds // 1000:03}"
        return formatted_time



def check_for_long_path():
    """If OS is Window's we'll check if long path is enabled"""
    import ctypes

    ntdll = ctypes.WinDLL("ntdll")

    if hasattr(ntdll, "RtlAreLongPathsEnabled"):
        ntdll.RtlAreLongPathsEnabled.restype = ctypes.c_ubyte
        ntdll.RtlAreLongPathsEnabled.argtypes = ()

        def are_long_paths_enabled():
            return bool(ntdll.RtlAreLongPathsEnabled())

        # call the function to determine if long paths are enabled
        return are_long_paths_enabled()

    else:
        return False



import platform
from pathlib import Path
from typing import Optional, Union

from pymediainfo import MediaInfo

from auto_qpf.generate_chapters import ChapterGenerator
from auto_qpf.long_path import check_for_long_path
from auto_qpf.qpf_exceptions import (
    ChapterIndexError,
    ImproperChapterError,
    NoChapterDataError,
)


class QpfGenerator:
    """
    Generates QPF file, accepted formats are media or OGM text.

    Public method: generate_qpf
    """

    def generate_qpf(
        self,
        file_input: Union[Path, str],
        file_output: Optional[Union[Path, str]] = None,
        write_to_disk: bool = True,
        fps: Union[int, float] = 23.976,
        auto_detect_fps: bool = True,
        generate_chapters: bool = True,
        generated_chapter_chunks: float = 5.0,
    ):
        """Creates chapter QPF and returns the path to the output file.

        Args:
            file_input (Union[Path, str]): File input.
            file_output (Union[Path, str], optional): File output. Defaults to None (will auto create name).
            write_to_disk (bool, optional): If this is set, writes to disk, if not it'll return an object with the qpf.
            fps (Union[int, float], optional): Source FPS. Defaults to 23.976. If auto_detect_fps is true and
            input is a not a .txt file, this will be over-ridden automatically.
            auto_detect_fps (bool, optional): Auto detects non text based sources if a video track is present.
            Defaults to True.
            generate_chapters (bool, optional): When enabled this will run a helper class to extract and/or generate new
            new chapters as needed to create a qpf code for.
            generated_chapter_chunks (float, optional): Defaults to 5.0 (percentage of input duration), if chapters
            are generated they will be generated every 'generated_chapter_chunks' percent.
        """
        # check if we're on a windows platform, if so check for long path support
        if platform.system() == "Windows":
            long_path = check_for_long_path()
            if not long_path:
                print(
                    "WARNING: Long path is not enabled for this OS. This can cause issues for paths greater > 260 characters."
                )

        # convert file_input to Path object
        file_input = Path(file_input)

        # auto generate file output if none was provided
        if not file_output:
            file_output = self._auto_output(file_input)
        else:
            file_output = Path(file_output)

        # check if input is a text file vs a media file
        detect_fps = None
        if file_input.suffix == ".txt":
            auto_detect_fps = False
            time_codes = self._get_time_codes_text(file_input)
        else:
            media_info = MediaInfo.parse(file_input, parse_speed=0.1)
            detect_fps = self._get_fps(media_info)

            if generate_chapters:
                # generate new output with _chapters before the extension for txt files
                txt_output = file_output.with_suffix("").with_name(
                    file_output.stem + "_chapters.txt"
                )

                new_file = ChapterGenerator().generate_ogm_chapters(
                    media_info_obj=media_info,
                    extract_tagged=False,
                    output_path=txt_output,
                    chapter_chunks=generated_chapter_chunks,
                )
                if not isinstance(new_file, Path):
                    raise NoChapterDataError(
                        "Chapter generation failed, cannot continue QPF generation."
                    )
                time_codes = self._get_time_codes_text(new_file)
            else:
                time_codes = self._get_time_codes_media_file(media_info)

        # if we're auto detecting fps attempt to update fps
        if auto_detect_fps:
            if detect_fps:
                fps = detect_fps

        # if no time codes are detected
        if not time_codes:
            raise NoChapterDataError("Could not detect chapter data from input")

        # get qpf list
        qpf_list = self._process_time_codes(time_codes, fps)

        # if we're writing to disk return the path to the output
        if write_to_disk:
            return self._write_qpf(file_output, qpf_list)

        # if we're not return the list of generated qpf frame positions
        else:
            return qpf_list

    def _process_time_codes(self, time_codes: list, fps: Union[float, str]):
        converted_time_codes = []
        for code in time_codes:
            convert_time = self._calculate_frame_position(code, fps)
            converted_time_codes.append(convert_time)
        return converted_time_codes

    def _get_time_codes_media_file(self, media_info: MediaInfo):
        time_code_list = []
        menu_stream_count = media_info.general_tracks[0].count_of_menu_streams
        if menu_stream_count and int(menu_stream_count) > 0:
            chapter_data = media_info.menu_tracks[0].to_data()
            if chapter_data:
                chapter_data = list(chapter_data.keys())
                try:
                    get_index = chapter_data.index("chapters_pos_end") + 1
                    split_list = chapter_data[get_index:]
                    for time_code in split_list:
                        time_code_list.append(time_code.replace("_", ":"))
                except ValueError:
                    raise ChapterIndexError(
                        "Cannot find the position of chapters_pos_end"
                    )
            else:
                raise NoChapterDataError("Input file has no chapter data")
        else:
            raise NoChapterDataError("Input file has no chapter data")

        return time_code_list

    @staticmethod
    def _auto_output(file_input: Path):
        return Path(file_input.parent / Path(file_input.name).with_suffix(".qpf"))

    @staticmethod
    def _get_time_codes_text(file_input: Path):
        time_code_list = []
        with open(file_input, "rt", encoding="utf-8") as txt_file:
            time_codes = txt_file.readlines()
            for num, chap in enumerate(time_codes):
                if num % 2 == 0:
                    time_code = chap.split("=")
                    if len(time_code) == 2:
                        time_code = time_code[1].replace("\n", "")
                        time_code_list.append(time_code)
                    else:
                        raise ImproperChapterError(
                            "Input file is an improper or corrupt chapter file."
                        )
        return time_code_list

    @staticmethod
    def _get_fps(media_info: MediaInfo):
        fps = None
        video_track_count = media_info.general_tracks[0].count_of_video_streams
        if video_track_count and int(video_track_count) > 0:
            fps = media_info.video_tracks[0].frame_rate
            if fps:
                if "." in fps:
                    fps = float(fps)
                else:
                    fps = int(fps)
        return fps

    @staticmethod
    def _calculate_frame_position(time_code: str, fps):
        # Formula: FramePosition = time-code (seconds) * fps
        # Split the time-code into hours, minutes, seconds, and milliseconds
        hours, minutes, seconds = map(float, time_code.split(":"))
        seconds += (hours * 3600) + (minutes * 60)

        # Calculate the frame position
        # We round to handle the floating point values as this is how it's done in MeGUI
        frame_position = round(seconds * fps)
        return frame_position

    @staticmethod
    def _write_qpf(output: Path, qpf_list: list):
        with open(output, "wt", encoding="utf-8") as qpf_file:
            for qpf_code in qpf_list:
                qpf_file.write(f"{qpf_code} K\n")
        return output



class ImproperChapterError(Exception):
    """Exception for improper chapter files"""


class NoChapterDataError(Exception):
    """No chapter data error"""


class ChapterIndexError(Exception):
    """Chapter index error"""



from .qpf import ChapterGenerator as ChapterGenerator
from .qpf import QpfGenerator as QpfGenerator
from .qpf_exceptions import ChapterIndexError as ChapterIndexError
from .qpf_exceptions import ImproperChapterError as ImproperChapterError
from .qpf_exceptions import NoChapterDataError as NoChapterDataError
